#include <math.h>
#include <memory.h>
#include "Noise3.h"

void Noise3::copy(const Noise3 &noise)
{
  l = noise.l;
  p = new Noise2 *[l];

  for (int i = 0; i < l; ++i) {
    p[i] = new Noise2;
    *p[i] = *noise.p[i];
  }
}

Noise3::Noise3()
{
  p = 0;
}

Noise3::Noise3(int n, int m, int l)
{
  this->l = l;
  p = new Noise2 *[l];

  for (int i = 0; i < l; ++i)
    p[i] = new Noise2(n, m);
}

Noise3::Noise3(const Noise3 &noise)
{
  copy(noise);
}

Noise3::~Noise3()
{
  if (p) {
    for (int i = 0; i < l; ++i)
      delete p[i];
    delete[] p;
  }
}

Noise3 &Noise3::operator=(const Noise3 &noise)
{
  this->~Noise3();
  copy(noise);
  
  return *this;
}

double Noise3::point(int i, int j, int k)
{
  return p[k]->point(i, j);
}

double Noise3::noise(double x, double y, double z)
{
  double lz = (double)l * (z - floor(z));
  int i = (int)lz;
  double t = lz - (double)i;

  if (i == 0) {
    return catmull_rom(p[l - 1]->noise(x, y), p[0]->noise(x, y),
      p[1]->noise(x, y), p[2]->noise(x, y), t);
  }
  else if (i < l - 2) {
    return catmull_rom(p[i - 1]->noise(x, y), p[i]->noise(x, y),
      p[i + 1]->noise(x, y), p[i + 2]->noise(x, y), t);
  }
  else if (i == l - 2) {
    return catmull_rom(p[l - 3]->noise(x, y), p[l - 2]->noise(x, y),
      p[l - 1]->noise(x, y), p[0]->noise(x, y), t);
  }
  else {
    return catmull_rom(p[l - 2]->noise(x, y), p[l - 1]->noise(x, y),
      p[0]->noise(x, y), p[1]->noise(x, y), t);
  }
}

double Noise3::perlin(double x, double y, double z, int o, double a)
{
  double v = noise(x, y, z);
  double w = a;

  while (--o > 0) {
    v += noise(x *= 2.0, y *= 2.0, z *= 2.0) * w;
    w *= a;
  }

  return v;
}

double Noise3::turbulence(double x, double y, double z, int o, double a)
{
  double v = fabs(noise(x, y, z));
  double w = a;

  while (--o > 0) {
    v += fabs(noise(x *= 2.0, y *= 2.0, z *= 2.0)) * w;
    w *= a;
  }

  return v;
}
